 //------------------------------------------------------------------
 // Copyright(c) 2022-2024 a2n Technology
 // Anwar Minarso (anwar.minarso@gmail.com)
 // https://github.com/anwarminarso/
 // This file is part of the a2n ESPWiFiMqttWrapper v1.0.6
 //
 // This library is free software; you can redistribute it and/or
 // modify it under the terms of the GNU Lesser General Public
 // License as published by the Free Software Foundation; either
 // version 2.1 of the License, or (at your option) any later version.
 //
 // This library is distributed in the hope that it will be useful,
 // but WITHOUT ANY WARRANTY; without even the implied warranty of
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
 // Lesser General Public License for more details
 //------------------------------------------------------------------


#ifndef ESPWiFiMqttWrapper_H
#define ESPWiFiMqttWrapper_H

//PubSubClient (require install)
//https://www.arduino.cc/reference/en/libraries/pubsubclient/
#include <PubSubClient.h>

// ESP32 Built In library
#if defined(ESP8266)
#include <ESP8266WiFi.h>
#elif defined(ESP32)
#include <WiFi.h>
#include <WiFiClientSecure.h>
#else
#error "This library only supports boards with ESP8266 or ESP32"
#endif

typedef std::function<void(char*, uint8_t*, unsigned int)> ArSubscribeHandlerFunction;
typedef std::function<void(const char*)> ArSubscribeMessageHandlerFunction;
typedef std::function<String()> ArPublishHandlerFunction;

template <typename T>
class ListOfNode {
	T _value;
public:
	ListOfNode<T>* next;
	ListOfNode(const T val) : _value(val), next(nullptr) {}
	~ListOfNode() {}
	const T& value() const { return _value; };
	T& value() { return _value; }
};
template <typename T, template<typename> class Item = ListOfNode>
class ListOf {
public:
	typedef Item<T> ItemType;
	typedef std::function<void(const T&)> OnRemove;
	typedef std::function<bool(const T&)> Predicate;
private:
	ItemType* _root;
	OnRemove _onRemove;

	class Iterator {
		ItemType* _node;
	public:
		Iterator(ItemType* current = nullptr) : _node(current) {}
		Iterator(const Iterator& i) : _node(i._node) {}
		Iterator& operator ++() { _node = _node->next; return *this; }
		bool operator != (const Iterator& i) const { return _node != i._node; }
		const T& operator * () const { return _node->value(); }
		const T* operator -> () const { return &_node->value(); }
	};

public:
	typedef const Iterator ConstIterator;
	ConstIterator begin() const { return ConstIterator(_root); }
	ConstIterator end() const { return ConstIterator(nullptr); }

	ListOf(OnRemove onRemove) : _root(nullptr), _onRemove(onRemove) {}
	~ListOf() {}
	void add(const T& t) {
		auto it = new ItemType(t);
		if (!_root) {
			_root = it;
		}
		else {
			auto i = _root;
			while (i->next) i = i->next;
			i->next = it;
		}
	}
	T& front() const {
		return _root->value();
	}

	bool isEmpty() const {
		return _root == nullptr;
	}
	size_t length() const {
		size_t i = 0;
		auto it = _root;
		while (it) {
			i++;
			it = it->next;
		}
		return i;
	}
	size_t count_if(Predicate predicate) const {
		size_t i = 0;
		auto it = _root;
		while (it) {
			if (!predicate) {
				i++;
			}
			else if (predicate(it->value())) {
				i++;
			}
			it = it->next;
		}
		return i;
	}
	const T* nth(size_t N) const {
		size_t i = 0;
		auto it = _root;
		while (it) {
			if (i++ == N)
				return &(it->value());
			it = it->next;
		}
		return nullptr;
	}
	bool remove(const T& t) {
		auto it = _root;
		auto pit = _root;
		while (it) {
			if (it->value() == t) {
				if (it == _root) {
					_root = _root->next;
				}
				else {
					pit->next = it->next;
				}

				if (_onRemove) {
					_onRemove(it->value());
				}

				delete it;
				return true;
			}
			pit = it;
			it = it->next;
		}
		return false;
	}
	bool remove_first(Predicate predicate) {
		auto it = _root;
		auto pit = _root;
		while (it) {
			if (predicate(it->value())) {
				if (it == _root) {
					_root = _root->next;
				}
				else {
					pit->next = it->next;
				}
				if (_onRemove) {
					_onRemove(it->value());
				}
				delete it;
				return true;
			}
			pit = it;
			it = it->next;
		}
		return false;
	}

	void free() {
		while (_root != nullptr) {
			auto it = _root;
			_root = _root->next;
			if (_onRemove) {
				_onRemove(it->value());
			}
			delete it;
		}
		_root = nullptr;
	}
};

class SubscribeHandler {
protected:
	const char* _topicFilter;
	ArSubscribeMessageHandlerFunction _func1;
	ArSubscribeHandlerFunction _func2;
public:
	const char* getTopicFilter() {
		return _topicFilter;
	}
	void setTopicFilter(const char* topicFilter) { _topicFilter = topicFilter; }
	void setFunction(ArSubscribeMessageHandlerFunction func) { _func1 = func; }
	void setFunction(ArSubscribeHandlerFunction func) { _func2 = func; }
	bool canHandle(const char* topic) {
		return strcmp(topic, _topicFilter) == 0;
	}
	void handleFunction(char* topic, uint8_t* payload, unsigned int length) {
		if (_func1) {
			String message;
			for (int i = 0; i < length; i++)
				message += (char)payload[i];
			_func1(message.c_str());
		}
		else if (_func2)
			_func2(topic, payload, length);
	}
};

class PublishHandler {
protected:
	const char* _topic;
	long _startDelay = 0;
	long _interval = 1000;
	long _delta = 0;
	uint32_t _lastMillis = 0;
	ArPublishHandlerFunction _func;
public:
	void setTopic(const char* topic) { _topic = topic; }
	const char* getTopic() {
		return _topic;
	}
	bool isTopicEqual(const char* topic) {
		return strcmp(topic, _topic) == 0;
	}
	void setFunction(ArPublishHandlerFunction func) { _func = func; }
	void setInterval(long interval) { _interval = interval; }
	void setStartDelay(long startDelay) { _startDelay = startDelay; }
	bool canHandle(uint32_t now) {
		if (_startDelay > now)
			return false;
		if (_lastMillis > now)
			_lastMillis = 0;
		_delta = now - _lastMillis;
		return _interval <= _delta;
	}
	const char * handleFunction(void) {
		char* result = nullptr;
		if (_func) {
			String val = _func();
			int str_len = val.length() + 1;
			result = (char*)malloc(str_len);
			val.toCharArray(result, str_len);
		}
		_lastMillis = millis();
		return result;
	}
};

class ESPWiFiMqttWrapper {
private:
	const char* _mqttServer = "iot.a2n.tech";
	uint16_t _mqttPort = 1883;

	WiFiClient _defaultClient;
	WiFiClientSecure _secureClient;

	PubSubClient _mqttClient;
	ListOf<SubscribeHandler*> _subscribehandlers;
	ListOf<PublishHandler*> _publishHandlers;
	Stream* _debugger;

	int _reconnectMqttCount = 0;
	int _reconnectWifiCount = 0;
	int _lastReconnect = 0;
	unsigned long now;
	int _maxReconnect = 30;
	bool _debug;
	bool _useSecureWiFi = false;


	const char* _mqttUsername;
	const char* _mqttPassword;
	const char* _mqttClientId = "";
	const char* _wifiHostName;
	const char* _wifiSSID;
	const char* _wifiPass;

	bool connectMqtt();
	bool connectWiFi();

	SubscribeHandler& addSubscribeHandler(SubscribeHandler* handler) {
		_subscribehandlers.add(handler);
		return *handler;
	};
	bool removeSubscribeHandler(SubscribeHandler* handler) {
		return _subscribehandlers.remove(handler);
	};
	PublishHandler& addPublishHandler(PublishHandler* handler) {
		_publishHandlers.add(handler);
		return *handler;
	};
	bool removePublishHandler(PublishHandler* handler) {
		return _publishHandlers.remove(handler);
	};
	void print(const String& s) {
		if (!_debug)
			return;
		_debugger->print(s);
	};
	void print(const char* c) {
		if (!_debug)
			return;
		_debugger->print(c);
	};
	void print(char c) {
		if (!_debug)
			return;
		_debugger->print(c);
	};
	void println(const String& s) {
		if (!_debug)
			return;
		_debugger->println(s);
	};
	void println(char c) {
		if (!_debug)
			return;
		_debugger->println(c);
	};
	void println(const char* c) {
		if (!_debug)
			return;
		_debugger->println(c);
	};
	void println(const Printable& p) {
		if (!_debug)
			return;
		_debugger->println(p);
	};
	void println(void) {
		if (!_debug)
			return;
		_debugger->println();
	};
	void setMqttServer();
#if defined(ESP8266)
	void setClock();
#endif
public:
	ESPWiFiMqttWrapper();
	void setMqttClientId(const char* ClientId);
	void setMqttServer(const char* mqttServer);
	void setMqttServer(const char* UserName, const char* Password);
	void setMqttServer(const char* mqttServer, uint16_t mqttPort);
	void setMqttServer(const char* mqttServer, const char* UserName, const char* Password);
	void setMqttServer(const char* mqttServer, uint16_t mqttPort, const char* UserName, const char* Password);
	void setDebugger(Stream* debugger) {
		_debug = true;
		_debugger = debugger;
	}
	SubscribeHandler& setSubscription(const char* topicFilter, ArSubscribeHandlerFunction func);
	SubscribeHandler& setSubscription(const char* topicFilter, ArSubscribeMessageHandlerFunction func);
	PublishHandler& setPublisher(const char* topic, int interval, ArPublishHandlerFunction func);
	PublishHandler& setPublisher(const char* topic, int interval, int startDelay, ArPublishHandlerFunction func);
	void removePublisher(const char* topic);
	void removeSubscription(const char* topicFilter);
	void setMaxReconnect(int value) {
		_maxReconnect = value;
	};
	void setWiFi(const char* hostName, const char* SSID, const char* wifiPassword);
#if defined(ESP32)
	void setCACert(const char* certificate);
	void setCertificate(const char* certificate);
	void setPrivateKey(const char* PrivateKey);
#elif defined (ESP8266)
	void setCACert(const X509List* certificate);
	void setCertificate(const X509List* certificate, const  PrivateKey* privateKey);
	//
#endif

	void useSecureWiFi(bool value) {
		_useSecureWiFi = value;
	}
	bool IsSecureWiFi() {
		return _useSecureWiFi;
	}
	void setWiFi(WiFiClient& client) {
		_defaultClient = client;
	}
	void setWiFiSecure(WiFiClientSecure& client) {
		_secureClient = client;
	}

	void initWiFi();
	void initMqtt();
	bool loop();
	PubSubClient getMqttClient() {
		return _mqttClient;
	};
	///* Overloaded functions end */
	//void printData();
	bool publish(const char* topic, const char* payload) {
		return _mqttClient.publish(topic, payload);
	}
	bool publish(const char* topic, const char* payload, boolean retained) {
		return _mqttClient.publish(topic, payload, retained);
	}
	bool publish(const char* topic, const uint8_t* payload, unsigned int plength) {
		return _mqttClient.publish(topic, payload, plength);
	}
	bool publish(const char* topic, const uint8_t* payload, unsigned int plength, boolean retained) {
		return _mqttClient.publish(topic, payload, plength, retained);
	}
	bool publish_P(const char* topic, const char* payload, boolean retained) {
		return _mqttClient.publish_P(topic, payload, retained);
	}
	bool publish_P(const char* topic, const uint8_t* payload, unsigned int plength, boolean retained) {
		return _mqttClient.publish_P(topic, payload, plength, retained);
	}
};
#endif
